'''Handling of old formats.'''


def legacy_load(obj):
    '''Handle legacy state formats and update to new format.'''
    version = obj.get('version')
    if not version:
        obj['version'] = '1.0.0'
        values = obj.get('attributes', {}).get('ssh_config')
        if values:
            obj['attributes']['ssh_config'] = {}
            for index, key in enumerate(
                    ('host', 'port', 'username', 'password')):
                try:
                    obj['attributes']['ssh_config'][key] = values[index]
                except IndexError:
                    obj['attributes']['ssh_config'][key] = ''
    return obj



########################################################################
# Acid Vault                                                           #
#                                                                      #
# This program is free software: you can redistribute it and/or modify #
# it under the terms of the GNU Affero General Public License as       #
# published by the Free Software Foundation, either version 3 of the   #
# License, or (at your option) any later version.                      #
#                                                                      #
# This program is distributed in the hope that it will be useful,      #
# but WITHOUT ANY WARRANTY; without even the implied warranty of       #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        #
# GNU Affero General Public License for more details.                  #
#                                                                      #
# You should have received a copy of the GNU Affero General Public     #
# License along with this program.  If not, see                        #
# <http://www.gnu.org/licenses/>.                                      #
########################################################################
'''Varius GUI widgets'''
import datetime
from distutils.version import LooseVersion
import json
import os
import subprocess
import sys
import tkinter
import urllib.request
import uuid

from version import __version__, __author__, __email__  # noqa:F401,E501 These are actually used
from version import __license__, __uri__, __summary__  # noqa:F401,E501 These are actually used
from vault import generate_password
from vault import VALID_PASSWORD_TYPES


class Dialog(tkinter.Toplevel):
    """
    Parent Dialog frame, inherit and override
    body and apply methods.
    """
    def __init__(self, parent, title=None, initial_data=None):
        super().__init__(parent)
        self.transient(parent)

        if title:
            self.title(title)

        self.parent = parent

        self.result = None

        body = tkinter.ttk.Frame(self)
        self.initial_focus = self.body(body, initial_data)
        body.pack(padx=5, pady=5, fill=tkinter.BOTH)

        self.buttonbox()

        self.grab_set()

        if not self.initial_focus:
            self.initial_focus = self

        self.protocol("WM_DELETE_WINDOW", self.cancel)

        self.geometry("+%d+%d" % (parent.winfo_rootx()+50,
                                  parent.winfo_rooty()+50))

        self.initial_focus.focus_set()

        self.wait_window(self)

    def body(self, master, initial_data=None):
        """Body of dialog, override in child classes."""
        pass

    def buttonbox(self):
        """Standard OK and Cancel buttons."""
        box = tkinter.ttk.Frame(self)

        w = tkinter.ttk.Button(
            box, text="OK", width=10, command=self.ok, default=tkinter.ACTIVE)
        w.pack(side='left', padx=5, pady=5)
        w = tkinter.ttk.Button(
            box, text="Cancel", width=10, command=self.cancel)
        w.pack(side='left', padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()
        return box

    def ok(self, event=None):
        """On OK button press."""
        if not self.validate():
            self.initial_focus.focus_set()
            return

        self.withdraw()
        self.update_idletasks()

        self.apply()

        self.cancel()

    def cancel(self, event=None):
        """On Cancel button press."""
        self.parent.focus_set()
        self.destroy()

    def validate(self):
        """
        Validate action on close, defaults to True if not overridden.
        """
        return 1

    def apply(self):
        """Apply result on OK button press, override in child class."""
        pass


class AddPassword(Dialog):
    '''Add/Edit password dialog.'''
    def body(self, master, initial_data):
        self.timer = Timer(master, self.close, 5000*60)
        master.bind_all('<Enter>', self.timer.reset)
        """Body of set key dialog."""
        self.initial_data = initial_data
        if initial_data and len(initial_data) == 6:
            self.uid = uuid.UUID(initial_data[0])
            start = 2
        else:
            self.uid = uuid.uuid4()
            start = 0
        for index, key in enumerate(('system',
                                     'username',
                                     'password',
                                     'notes'),
                                    start=start):
            setattr(self, key, tkinter.StringVar())
            e = LabelEntry(master,
                           width=50,
                           label=key.title(),
                           textvariable=getattr(self, key))
            e.pack()
            if initial_data:
                getattr(self, key).set(initial_data[index])

        f = tkinter.Frame(master)
        f.pack()
        pw_type = tkinter.StringVar()
        pw_type.set(VALID_PASSWORD_TYPES[0])
        pw_len = tkinter.IntVar()
        pw_len.set(10)
        gen_pass = tkinter.OptionMenu(f, pw_type, *VALID_PASSWORD_TYPES)
        gen_pass.configure(width=20)
        length = tkinter.OptionMenu(f, pw_len, *range(4, 16))
        length.configure(width=2)

        gen_pass.pack(side='left')
        length.pack(side='left')
        b = tkinter.Button(
            f,
            text='Generate Password',
            command=lambda: self.password.set(
                generate_password(
                    pw_type.get(), pw_len.get())))
        b.pack(side='left')

    def apply(self):
        """Set result upon OK button press."""
        # Only return result if it has changed.
        result = (self.uid,
                  datetime.datetime.utcnow(),
                  self.system.get(),
                  self.username.get(),
                  self.password.get(),
                  self.notes.get())
        if self.initial_data and len(self.inital_data) == len(result):
            if [index for index in range(2, 6) if
                    result[index] != self.initial_data[index]]:
                self.result = result
        else:
            self.result = result

    def close(self):
        self.timer.stop()
        self.destroy()


class SetupSSH(Dialog):
    '''Setup SSH related settings.'''
    def body(self, master, initial_data):
        """Body of SSH settings dialog."""
        self.clear_on_exit = tkinter.BooleanVar()
        self.clear_on_exit.set(initial_data.get('clear_on_exit', True))
        clear_on_exit = tkinter.Checkbutton(master,
                                            text='Clear SSH settings on exit',
                                            variable=self.clear_on_exit,
                                            anchor='w')
        clear_on_exit.pack(expand=1, fill=tkinter.X)
        for key in ('host', 'port', 'username', 'password'):
            setattr(self, key, tkinter.StringVar())
            e = LabelEntry(master,
                           width=50,
                           label=key.title().replace('_', ' '),
                           textvariable=getattr(self, key))
            e.pack()
            if key == 'password':
                e.set_config(show='*')
            if initial_data:
                try:
                    getattr(self, key).set(initial_data[key])
                except KeyError:
                    pass

    def apply(self):
        """Set result upon OK button press."""
        self.result = {key: getattr(self, key).get() for
                       key in ('host', 'port', 'username',
                               'password', 'clear_on_exit')}


class SetupFiles(Dialog):
    '''File related settings, paths and steganography.'''
    def body(self, master, initial_data):
        # Checkboxes.
        for key, default in (('use_steganography', False),
                             ('clear_on_exit', True),
                             ('sync', True)):
            value = initial_data.get(key, default)
            setattr(self, key, tkinter.BooleanVar(
                self, name=key, value=value))
            c = tkinter.Checkbutton(
                master,
                text=key.replace('_', ' ').title(),
                variable=getattr(self, key),
                anchor='w')
            c.pack(expand=1, fill=tkinter.X)

        # Entries
        for key, name in (('file_location',
                           'Password file location'),
                          ('original_file',
                           'Steganography original file location')):
            setattr(self, key, tkinter.StringVar())
            if initial_data:
                getattr(self, key).set(initial_data.get(key, ''))

            e = LabelEntry(master,
                           width=50,
                           label=name,
                           textvariable=getattr(self, key))
            e.pack()

    def apply(self):
        self.result = {key: getattr(self, key).get() for
                       key in ('sync', 'file_location', 'original_file',
                               'use_steganography', 'clear_on_exit')}


class About(Dialog):
    '''About information and update program button.'''
    def body(self, master, _):
        for name, var in (('', '__summary__'),
                          ('Version: ', '__version__'),
                          ('Author: ', '__author__'),
                          ('Contact: ', '__email__'),
                          ('Licence: ', '__license__'),
                          ('', '__uri__')):
            label = tkinter.Label(master, text=f'{name} {eval(var)}')
            label.pack(fill=tkinter.X, expand=1)

        version = check_version('acid_vault')
        if version != __version__:
            label = tkinter.Label(
                master, text=f'Version on PyPi: {version}')
            label.pack(fill=tkinter.X, expand=1)
            button = tkinter.Button(master,
                                    text='Update (Will restart program)',
                                    command=lambda: self.update(version))
            button.pack()

    def update(self, version):
        '''Update to newest version of program'''
        p = subprocess.Popen(
            [sys.executable, '-m', 'pip', 'install', f'acid_vault=={version}'],
            stdout=subprocess.PIPE)
        p.wait()
        print(str(p.stdout.read(), 'utf-8'))
        path = os.path.join(os.path.dirname(__file__), 'VaultGui.pyw')
        print(f'Starting path: {path}')
        os.execv(sys.executable, [sys.executable, path])

    def buttonbox(self):
        """Standard OK and Cancel buttons."""
        box = tkinter.ttk.Frame(self)

        w = tkinter.ttk.Button(
            box, text="OK", width=10, command=self.ok, default=tkinter.ACTIVE)
        w.pack(side='left', padx=5, pady=5)
        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()
        return box


class Box(tkinter.Frame):
    """Frame box to make widgets align."""
    def __init__(self, master, widget_type, *args, **kwargs):
        self.master = master
        side = kwargs.pop('side', 'bottom')
        super().__init__(master)
        self.widget = getattr(tkinter, widget_type)(self, *args, **kwargs)
        self.widget.pack(side=side)

    def config(self, *args, **kwargs):
        self.widget.config(*args, **kwargs)


class StatusBar(tkinter.Label):
    '''Status bar widget.'''
    def __init__(self, master, *args, **kwargs):
        self.default_color = master.cget('bg')
        self.master = master
        self._data = kwargs.pop('textvariable', tkinter.StringVar())
        super().__init__(master, *args, textvariable=self._data, **kwargs)

    def set(self, data, color=None):
        color = color or self.default_color
        self._data.set(data)
        self.config(bg=color)
        self.master.update_idletasks()


class LabelEntry(tkinter.Frame):
    '''A labled entry frame.'''
    def __init__(self, master, *args, **kwargs):
        self.master = master
        label = kwargs.pop('label', '')
        super().__init__(master)
        self._label = tkinter.Label(self, text=label)
        self._entry = tkinter.Entry(self, *args, **kwargs)
        self._label.pack(side='top', fill=tkinter.X)
        self._entry.pack(side='top', fill=tkinter.X)

    @property
    def label(self, value):
        self._label.set(value)

    def set_config(self, **kwargs):
        self._entry.configure(**kwargs)

    def focus_set(self, *args, **kwargs):
        self._entry.focus_set(*args, **kwargs)

    def bind(self, *args, **kwargs):
        self._entry.bind(*args, **kwargs)


class Timer():
    """Timer class to triger call back after given time."""
    def __init__(self, master, callback, after, reset_after_trigger=False,
                 *args, **kwargs):
        self.master = master
        self.callback = callback
        self.after = after
        self.args = args
        self.kwargs = kwargs
        self.reset_after_trigger = reset_after_trigger
        self.timer = master.after(self.after, self._trigger)

    def _trigger(self):
        self.callback(*self.args, **self.kwargs)
        if self.reset_after_trigger:
            self.timer = self.master.after(self.after, self._trigger)

    def reset(self, *args, **kwargs):
        """Reset timer"""
        self.timer and self.master.after_cancel(self.timer)
        self.timer = self.master.after(self.after, self._trigger)

    def stop(self):
        """Stop timer."""
        self.timer and self.master.after_cancel(self.timer)

    def start(self):
        """Start timer, only used after it has been stoped."""
        self.timer = self.master.after(self.after, self._trigger)


def check_version(name):
    '''Checks current version for package on pypi'''
    pypi_url = f'https://pypi.org/pypi/{name}/json'
    response = urllib.request.urlopen(pypi_url, timeout=5).read().decode()
    latest_version = max(LooseVersion(s) for s in
                         json.loads(response)['releases'].keys())
    return latest_version



###############################################################################
# Acid Vault                                                                  #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU Affero General Public License as published by #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU Affero General Public License for more details.                         #
#                                                                             #
# You should have received a copy of the GNU Affero General Public License    #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################
from . import *



########################################################################
# Acid Vault                                                           #
#                                                                      #
# This program is free software: you can redistribute it and/or modify #
# it under the terms of the GNU Affero General Public License as       #
# published by the Free Software Foundation, either version 3 of the   #
# License, or (at your option) any later version.                      #
#                                                                      #
# This program is distributed in the hope that it will be useful,      #
# but WITHOUT ANY WARRANTY; without even the implied warranty of       #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        #
# GNU Affero General Public License for more details.                  #
#                                                                      #
# You should have received a copy of the GNU Affero General Public     #
# License                                                              #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.#
########################################################################
'''
Encrypts and decrypts data.
'''
import ast
import base64
import datetime
import os
import pickle
import random
import string
import uuid

from cryptography.fernet import Fernet
from cryptography.fernet import InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

import ssh
import steganography
import version

VERSION = '1.0.0'
VALID_PASSWORD_TYPES = ('alpha',
                        'alphanum',
                        'alphanumspecial',
                        'mobilealpha',
                        'mobilealphanum',
                        'mobilealphanumspecial',
                        'numerical')


class VaultError(Exception):
    '''Vault releated errors.'''
    pass


class Vault():
    '''Class to hold salt, iterations and the data.'''
    def __init__(self,
                 data_path=None,
                 ssh_params=None,
                 path_to_original=None,
                 update=True):
        self._locked = False
        # When loaded from local file do not update.
        self.update = update

        if data_path:
            self.load(data_path, ssh_params, path_to_original)
        else:
            self.data = {'salt': os.urandom(16),
                         'iterations': 1000000,
                         'vault': [],
                         'timestamp': datetime.datetime.utcnow(),
                         'version': VERSION}

    @property
    def locked(self):
        '''Locked status of the vault'''
        return self._locked

    @property
    def timestamp(self):
        return self.data.get('timestamp')

    def _open(self, file_path, ssh_params, path_to_orginal, mode, call):
        if ssh_params:
            with ssh.RemoteFile(ssh_params, file_path, mode) as fh:
                if not fh:
                    raise VaultError('Could not aquire lock')
                return call(fh, path_to_orginal)
        else:
            with open(file_path, mode) as fh:
                return call(fh, path_to_orginal)

    def force_lock(self, ssh_params):
        ssh.force_lock(**ssh_params)

    def update_version(self, password):
        if not self.update:
            return
        if not version.is_greater_version(VERSION, self.data.get('version')):
            return
        # Save status of lock so we know if we should lock again.
        lock_status = self.locked
        if lock_status:
            self.unlock(password)
        for record in self.data['vault']:
            record['date'] = record.get('date', '')
            record['uid'] = record.get('uid', uuid.uuid4())
        if lock_status:
            self.lock(password)
        return True

    def check_remote(self, file_path, ssh_params, path_to_original):
        def check(fh, path_to_original):
            if path_to_original:
                data = pickle.loads(steganography.read(fh, path_to_original))
            else:
                data = pickle.load(fh)
            remote_ts = data.get('timestamp')
            remote_ver = data.get('version')
            local_ts = self.data.get('timestamp')
            local_ver = self.data.get('version')

            if not (remote_ver and
                    local_ver and
                    version.same_minor_version(remote_ver, local_ver)):
                raise VaultError(
                    f'Version missmatch: {remote_ver=}, {local_ver=}')
            if remote_ts and local_ts and remote_ts > local_ts:
                return data
        return self._open(file_path, ssh_params, path_to_original, 'rb', check)

    def merge(self, password, data, file_path, ssh_params, path_to_original):
        self.unlock(password)
        data = self._unlock(password, data)
        updated = False
        current = {obj['uid']: obj['date'] for obj in self.data['vault']}
        for obj in data:
            if 'uid' in obj and obj['uid'] not in current:
                self.add(obj)
                updated = True
            elif 'date' in obj and obj['date'] > current[obj['uid']]['date']:
                self.replace(obj)
                updated = True
        self.lock(password)
        if updated:
            self.save(file_path, ssh_params, path_to_original)

    def load(self, file_path, ssh_params=None, path_to_original=None):
        def read(fh, path_to_original):
            if path_to_original:
                self.data = pickle.loads(
                    steganography.read(fh, path_to_original))
            else:
                self.data = pickle.load(fh)
        self._open(file_path, ssh_params, path_to_original, 'rb', read)
        self._locked = True

    def save(self, file_path, ssh_params=None, path_to_original=None):
        def write(fh, path_to_orginal):
            if path_to_original:
                steganography.write(
                    fh, path_to_original, pickle.dumps(self.data))
            else:
                fh.write(pickle.dumps(self.data))
        self.data['timestamp'] = datetime.datetime.utcnow()
        self.data['version'] = VERSION
        self._open(file_path, ssh_params, path_to_original, 'wb', write)

    def load_clear(self, fh):
        '''Load data from open file containing clear text data.'''
        if self.locked:
            raise VaultError('Vault is locked, unlock first!')
        self._locked = False
        for row in fh:
            self.add(ast.literal_eval(row.strip()))

    def save_clear(self, fh):
        '''Save data in open file as clear text.'''
        if self.locked:
            raise VaultError('Vault is locked, please unlock first')
        for obj in self.data['vault']:
            print(repr(obj), file=fh)

    def lock(self, password):
        '''Lock vault with password.'''
        if self._locked:
            return
        data = pickle.dumps(self.data['vault'])
        key = self.create_key(password,
                              self.data['salt'],
                              self.data.get('iterations', 1000000))
        self.data['vault'] = Fernet(key).encrypt(data)
        self._locked = True

    def unlock(self, password):
        '''Unlock vault with password.'''
        if not self._locked:
            return
        data = self._unlock(password, self.data)
        if data:
            self.data = data
            self._locked = False

    def _unlock(self, password, data):
        key = self.create_key(password,
                              data['salt'],
                              data.get('iterations', 1000000))
        try:
            data['vault'] = pickle.loads(Fernet(key).decrypt(data['vault']))
            return data
        except InvalidToken:
            pass

    def create_key(self, password, salt, iterations=1000000):
        '''Create a key to be used.'''
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=iterations,
            backend=default_backend()
            )
        return base64.urlsafe_b64encode(
            kdf.derive(bytearray(password, 'utf-8')))

    def get_objects(self):
        '''Get objects in their current state in vault.'''
        return self.data['vault']

    def set_objects(self, objs):
        '''Set vault content to input.'''
        if not self.locked:
            self.data['vault'] = []
            for obj in objs:
                self.add(obj)

    def add(self, obj):
        '''Add to vault content.'''
        self.data['vault'].append(obj)
        self.data['timestamp'] = datetime.datetime.utcnow()

    def replace(self, obj):
        for index, current_obj in enumerate(self.data['vault']):
            if current_obj['uid'] == obj['uid']:
                self.data['vault'][index] = obj
                self.data['timestamp'] = datetime.datetime.utcnow()
                return True

    def remove_password(self, obj):
        '''Remove obj from vault if it exists.'''
        self.data['vault'].remove(obj)


def generate_password(password_type='alpha', n=10):
    '''Password generator to sugest randomized passwords.'''
    alphabets = {'alpha': (string.ascii_letters, 'isupper', 'islower'),
                 'alphanum': (string.ascii_letters + string.digits, 'isupper', 'islower', 'isdigit'),  # noqa: E501 Line to long
                 'alphanumspecial': (string.ascii_letters + string.digits + '!"#¤%&/()=?', 'isupper', 'islower', 'isdigit'),  # noqa: E501 Line to long
                 'mobilealpha': (string.ascii_lowercase, 'islower'),
                 'mobilealphanum': (string.ascii_lowercase, 'islower'),
                 'mobilealphanumspecial': (string.ascii_lowercase, 'islower'),
                 'numerical': (string.digits, 'isdigit')}
    if password_type not in alphabets:
        raise KeyError('Password type has to be one of the following:'
                       f' {", ".join(alphabets.keys())} it is'
                       f' "{password_type}".')
    while True:
        alphabet = alphabets[password_type][0]
        password = random.choices(alphabet, k=n)
        if password_type.startswith('mobile'):
            password[0] = password[0].upper()
        if 'mobilealphanum' in password_type:
            password[-2] = random.choice(string.digits)
        if password_type == 'mobilealphanumspecial':
            password[-1] = random.choice('!"#¤%&/()=?')

        for test in alphabets[password_type][1:]:
            if not [c for c in password if getattr(c, test)()]:
                break
        else:
            if password_type == 'alphanumspecial':
                if not [c for c in password if not c.isalpha()]:
                    continue
            break
    return ''.join(password)



########################################################################
# Acid Vault                                                           #
#                                                                      #
# This program is free software: you can redistribute it and/or modify #
# it under the terms of the GNU Affero General Public License as       #
# published by the Free Software Foundation, either version 3 of the   #
# License, or (at your option) any later version.                      #
#                                                                      #
# This program is distributed in the hope that it will be useful,      #
# but WITHOUT ANY WARRANTY; without even the implied warranty of       #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        #
# GNU Affero General Public License for more details.                  #
#                                                                      #
# You should have received a copy of the GNU Affero General Public     #
# License along with this program.  If not, see                        #
# <http://www.gnu.org/licenses/>.                                      #
########################################################################

'''
Steganography
Hides data within an image so that its not obvious that the data is
there.
'''

from PIL import Image


class SteganographyError(Exception):
    '''StganographyError for various error situations.'''
    pass


def write(fh, original, data):
    '''
    Combine data with data from original (image (tested on jpg))
    and write it as a .png image in opened file.
    '''
    with Image.open(original) as image:
        mask = [int(y) for y in ''.join([str(format(x, 'b')).zfill(8) for
                                         x in data])]
        bands = image.split()
        for band_index in (0, 1, 2):
            new_band = []
            for index, x in enumerate(image.getdata(band_index)):
                new_band.append((x + (mask.pop(0) if mask else 2) % 255))
            bands[band_index].putdata(new_band)
            if not mask:
                break
        else:
            raise SteganographyError('Ran out of image space')
        i = Image.merge(image.mode, bands)
        i.save(fh, 'png')


def read(fh, original):
    '''
    Read data from opened file and compare it to orignal to get stored data.
    '''
    def convert_result(result):
        return int(''.join(result), 2).to_bytes(len(result) // 8,
                                                byteorder='big')
    with Image.open(original) as orig, Image.open(fh) as mask:
        result = []
        for band_index in (0, 1, 2):
            orig_data = list(orig.getdata(band_index))
            mask_data = list(mask.getdata(band_index))
            for x, y in zip(mask_data, orig_data):
                value = x - y
                if value in (2, -254):
                    return convert_result(result)
                if value < 0:
                    value = 1
                result.append(str(value))
        return convert_result(result)



import hashlib
import PIL
import pytest

from tests.fixtures import *
import acid_vault.steganography as steganography
from acid_vault.steganography import SteganographyError

def test_write(tmpdir):
    expected = 'ce91ce323ab4c03f705bc37c5071af1c'
    data = b'Test string to write'
    file_path = tmpdir.join('outfile.png')
    original_path = os.path.join(os.path.dirname(__file__),
                                 "test_data",
                                 "test.jpg")
    with open(file_path, 'bw') as fh:
        steganography.write(fh, original_path, data)
    with open(file_path, 'br') as fh:
        assert hashlib.md5(fh.read()).hexdigest() == expected


def test_read():
    expected = b'Test string to write'
    original_path = os.path.join(os.path.dirname(__file__),
                                 "test_data",
                                 "test.jpg")
    written_path = os.path.join(os.path.dirname(__file__),
                                "test_data",
                                "test.png")
    with open(written_path, 'rb') as fh:
        assert steganography.read(fh, original_path) == expected

def test_to_large(tmpdir):
    file_path = tmpdir.join('outfile.png')
    original_path = os.path.join(os.path.dirname(__file__),
                                 "test_data",
                                 "test.jpg")
    with PIL.Image.open(original_path) as image:
        data = b'a'*len(image.getdata())
    with open(file_path, 'bw') as fh:
        with pytest.raises(SteganographyError):
            steganography.write(fh, original_path, data)


###############################################################################
# Acid Vault                                                                  #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU Affero General Public License as published by #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU Affero General Public License for more details.                         #
#                                                                             #
# You should have received a copy of the GNU Affero General Public License    #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################
"""
Handle files on Remote server through ssh.
Wrapper around Paramiko sftp server.
"""
import os
import paramiko
import tempfile
import tkinter.messagebox

import constants

LOCK_PATH = r'.lock/vault.lock'


class MissingKeyError(Exception):
    def __init__(self, client, hostname, key):
        super().__init__(f'{hostname}: {key.get_base64()} '
                         'is missing in hostfile')
        self.client = client
        self.hostname = hostname
        self.key = key


class RemoteFile():
    """
    Class to handle remote files through ssh.
    Implements context manager.
    """
    def __init__(self, ssh_params, filepath, *args, **kwargs):
        host, port, username, password = [ssh_params.get(x) for x in
                                          ('host',
                                           'port',
                                           'username',
                                           'password')]
        self.safe_write = None
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(RejectKeyPolicy())
        self.known_hosts = os.path.join(constants.data_dir(), '.know_hosts')
        try:
            self.ssh.load_host_keys(self.known_hosts)
        except FileNotFoundError:
            os.makedirs(constants.data_dir(), exist_ok=True)
            with open(self.known_hosts, 'w'):
                pass
            self.ssh.load_host_keys(self.known_hosts)
        self._connect(host, port, username, password)
        self.filepath = filepath
        self.sftp = None
        self.stat = None
        self.lock = None
        self.lock_script_path = '/tmp/lock_script.cmd'
        self.fh = self._open(*args, **kwargs)

    def __enter__(self):
        return self.fh

    def __exit__(self, *args):
        self.close()

    def _connect(self, host, port, username, password):
        try:
            self.ssh.connect(host, int(port), username, password)
        except paramiko.ssh_exception.BadHostKeyException as err:
            if tkinter.messagebox.askokcancel(
                    'Bad Host Key',
                    f'{err}\n '
                    'This may be a Man In the Middle attack\n '
                    'Continue anyway?'):
                keys = self.ssh.get_host_keys()
                hostname = (err.hostname if err.hostname in keys else
                            f'[{err.hostname}]:{port}')
                if hostname not in keys:
                    raise
                self.add_key(hostname, err.key)
                self._connect(host, port, username, password)
                return

        except paramiko.ssh_exception.SSHException as err:
            if str(err) == 'Authentication failed.':
                tkinter.messagebox.showerror('Authentication failed.',
                                             'Authentication failed.')
        except MissingKeyError as err:
            if tkinter.messagebox.askokcancel(
                    'Missing Host Key',
                    f'{err}\n '
                    'Continue anyway?'):
                self.add_key(err.hostname, err.key)
                self._connect(host, port, username, password)
                return
        else:
            # All OK
            return
        # If we have here an exception has happened
        self.ssh.close()

    def add_key(self, hostname, key):
        """Adds keys to host file."""
        keys = self.ssh.get_host_keys()
        keys.pop(hostname, '')
        keys.add(hostname, 'ssh-rsa', key)
        keys.save(self.known_hosts)

    def _makedirs(self, path):
        current_path = ''
        for p in path.split('/'):
            current_path = (current_path and f'{current_path}/{p}') or p
            try:
                self.sftp.stat(current_path)
            except FileNotFoundError:
                self.sftp.mkdir(current_path)

    def close(self):
        """Close all open handles in the correct order."""
        if self.safe_write:
            self.ssh.exec_command(f'mv {self.filepath}.bak {self.filepath}')
        if self.stat:
            self.sftp.utime(self.filepath, (self.stat.st_atime,
                                            self.stat.st_mtime))
        for con in ('fh', 'sftp', 'ssh'):
            try:
                getattr(self, con).close()
            except Exception:
                pass
        if self.lock:
            try:
                self.lock.write('quit')
            except Exception as error:
                print(f'Failed to close lock: {error}')

    def _open(self, mode='r', *args, timeout=15, **kwargs):
        """Open remote path."""
        path = kwargs.pop('path', self.filepath)
        transport = self.ssh.get_transport()
        try:
            transport.send_ignore()
        except Exception:
            return
        self.sftp = self.ssh.open_sftp()
        self._makedirs(os.path.dirname(path))
        try:
            self.stat = self.sftp.stat(path)
        except FileNotFoundError:
            pass
        if [x for x in args if 'w' in x] or 'w' in kwargs.get('mode', ''):
            options = f'-e -w {timeout}'
        else:
            options = f'-s -w {timeout}'
        self.lock = self.aquire_lock(options)
        if not self.lock:
            print('Failed to aquire lock')
            return
        if 'w' in mode and self.filepath == path:
            self.safe_write = True
            path = f'{path}.bak'
        return self.sftp.open(path, *args, mode=mode, **kwargs)

    def create_lock_script(self):
        try:
            self.sftp.stat(self.lock_script_path)
        except (IOError, FileNotFoundError):
            fh, file_path = tempfile.mkstemp(text=True)
            with self.sftp.open(self.lock_script_path, 'w') as fh:
                print('#!/bin/sh\n'
                      'echo "OK"\n'
                      'INPUT=init\n'
                      'while [ "$INPUT" != "quit" ]\n'
                      '    do\n'
                      '        read INPUT\n'
                      '    done\n',
                      file=fh)
            self.ssh.exec_command(f'chmod +x {self.lock_script_path}')

    def aquire_lock(self, options='-e -w 15'):
        self.create_lock_script()
        self.ssh.exec_command(
            f'mkdir -p {os.path.dirname(LOCK_PATH)}')
        stdin, stdout, stderr = self.ssh.exec_command(
            f'flock {options} {LOCK_PATH} -c {self.lock_script_path}')
        if stdout.readline():
            return stdin

    def release_lock(self, pipe):
        pipe.write('quit')


def force_lock(host, port, user, password=None):
    client = paramiko.SSHClient()
    client.connect(host, int(port), user, password)
    client.exec_command(f'rm {LOCK_PATH}')
    client.close()


def load_host_keys(client):
    data_dir = constants.data_dir()
    try:
        client.load_host_keys(os.path.join(data_dir, '.know_hosts'))
    except FileNotFoundError:
        os.makedirs(data_dir, exist_ok=True)
        with open(os.path.join(data_dir, '.know_hosts'), 'w'):
            pass


def upload_pub_key(
        host, port, user, password, key_path, accept_unknown_host=False,
        comment=False):
    client = paramiko.SSHClient()
    if accept_unknown_host:
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy)
    client.connect(host, int(port), user, password)
    transport = client.get_transport()
    transport.send_ignore()
    sftp = client.open_sftp()
    key = read_key(key_path)
    try:
        sftp.mkdir('.ssh')
    except IOError:
        pass
    with sftp.open('.ssh/authorized_keys', 'r+') as fh:
        existing_keys = [x.rstrip('\n') for x in fh.readlines()]
        if key in existing_keys:
            return
        if comment:
            print(f'# {comment}', file=fh)
        print(key, file=fh)
    sftp.close()
    transport.close()
    client.close()


def read_key(key_path):
    with open(key_path) as fh:
        rows = [x.strip('\n') for x in fh.readlines()]
        comment = rows[1].split('"')[1]
        key = ''.join(rows[2:-1])
        return f'ssh-rsa {key} {comment}'


class RejectKeyPolicy(paramiko.client.MissingHostKeyPolicy):
    def missing_host_key(self, client, hostname, key):
        raise MissingKeyError(client, hostname, key)



###############################################################################
# Acid Vault                                                                  #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU Affero General Public License as published by #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU Affero General Public License for more details.                         #
#                                                                             #
# You should have received a copy of the GNU Affero General Public License    #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################
'''
Constants.
'''

import os
import platform


def data_dir():
    '''Returns path to programs data dir.'''
    system = platform.system()
    if system == 'Windows':
        path = os.path.join('appdata', 'local', 'vault')
    elif system == 'Linux':
        path = '.vault'
    else:
        path = ''
    path = os.path.expanduser(os.path.join('~', path))
    os.makedirs(path, exist_ok=True)
    return path
